from typing import Any

from pytest_matchers.matchers import Matcher
from pytest_matchers.utils.matcher_utils import as_matcher


class Not(Matcher):
    def __init__(self, matcher: Matcher | Any):
        self._matcher = as_matcher(matcher)

    def __new__(cls, matcher):
        if isinstance(matcher, Not):
            return matcher._matcher
        return super().__new__(cls)

    def __invert__(self) -> Matcher:
        return self._matcher

    def matches(self, value) -> bool:
        return not self._matcher == value

    def __repr__(self) -> str:
        return f"To not be {self._matcher.concatenated_repr()}"



from .main import (
    anything,
    between,
    case,
    different_value,
    has_attribute,
    if_false,
    if_true,
    is_datetime,
    is_datetime_string,
    is_dict,
    is_instance,
    is_json,
    is_list,
    is_number,
    is_string,
    one_of,
    same_value,
)



from .base import Matcher
from .length import Length
from .contains import Contains
from .starts_with import StartsWith
from .ends_with import EndsWith
from .eq import Eq
from .between import Between

from .and_matcher import And
from .is_instance import IsInstance
from .is_list import IsList
from .is_string import IsString
from .or_matcher import Or
from .not_matcher import Not
from .is_number import IsNumber
from .anything import Anything
from .datetime import Datetime
from .datetime_string import DatetimeString
from .has_attribute import HasAttribute
from .same_value import SameValue
from .different_value import DifferentValue
from .if_matcher import If
from .case import Case
from .dict import Dict
from .json import JSON



from typing import Any

from pytest_matchers.matchers import Matcher
from pytest_matchers.utils.repr_utils import concat_reprs


class Or(Matcher):
    def __init__(self, *matchers: Matcher | Any):
        self._matchers = matchers

    def matches(self, value: Any) -> bool:
        return any(matcher == value for matcher in self._matchers)

    def __repr__(self) -> str:
        return concat_reprs("", repr(self._matchers[0]), *self._matchers[1:], separator="or")



import json
from typing import Any

from pytest_matchers.matchers import IsInstance, Matcher
from pytest_matchers.matchers.dict import dict_matcher
from pytest_matchers.utils.matcher_utils import matches_or_none
from pytest_matchers.utils.repr_utils import concat_reprs


class JSON(Matcher):
    def __init__(self, matching: dict = None, *, exclude: list = None):
        self._is_instance_matcher = IsInstance(str)
        self._dict_matcher = dict_matcher(matching, exclude)

    def matches(self, value: Any) -> bool:
        if self._is_instance_matcher == value:
            try:
                json_data = json.loads(value)
            except json.JSONDecodeError:
                return False
            return matches_or_none(self._dict_matcher, json_data)
        return False

    def __repr__(self) -> str:
        dict_repr = self._dict_matcher.expectations_repr() if self._dict_matcher else ""
        return concat_reprs("To be a JSON", dict_repr)



from typing import Any, Callable

from pytest_matchers.matchers import Anything, Matcher
from pytest_matchers.utils.matcher_utils import as_matcher
from pytest_matchers.utils.repr_utils import capitalized
from pytest_matchers.utils.warn import warn


class If(Matcher):
    def __init__(
        self,
        condition: Callable | bool | Any,
        then: Matcher | Any = None,
        or_else: Matcher | Any = None,
    ):
        self._condition = condition
        self._then = as_matcher(then or Anything())
        self._or_else = as_matcher(or_else or Anything())

    def matches(self, value: Any) -> bool:
        matcher = self._or_else
        if self._satisfies_condition(value):
            matcher = self._then
        return matcher == value

    def _satisfies_condition(self, value: Any) -> bool:
        if isinstance(self._condition, bool):
            return self._condition
        if not callable(self._condition):
            return self._condition == value
        try:
            return self._condition(value)
        except Exception as error:  # pylint: disable=broad-except
            warn(f"Condition failed with error: {error}. Returning False as default value.")
            return False

    def __repr__(self):
        if isinstance(self._condition, bool):
            if self._condition:
                return repr(self._then)
            return repr(self._or_else)

        representation = (
            f"{repr(self._then)} if condition returns True"
            f" else {self._or_else.concatenated_repr()}"
        )
        return capitalized(representation)



from typing import Any

from pytest_matchers.matchers import IsInstance, Matcher
from pytest_matchers.utils.matcher_utils import (
    contains_matcher,
    ends_with_matcher,
    length_matcher,
    matches_or_none,
    starts_with_matcher,
)
from pytest_matchers.utils.repr_utils import concat_reprs


class IsString(Matcher):
    def __init__(
        self,
        *,
        starts_with: str = None,
        ends_with: str = None,
        contains: str = None,
        length: int = None,
        max_length: int = None,
        min_length: int = None,
    ):
        self._starts_with_matcher = starts_with_matcher(starts_with)
        self._ends_with_matcher = ends_with_matcher(ends_with)
        self._contains_matcher = contains_matcher(contains)
        self._length_matcher = length_matcher(length, min_length, max_length)

    def matches(self, value: Any) -> bool:
        return (
            IsInstance(str).matches(value)
            and matches_or_none(self._length_matcher, value)
            and matches_or_none(self._contains_matcher, value)
            and matches_or_none(self._starts_with_matcher, value)
            and matches_or_none(self._ends_with_matcher, value)
        )

    def __repr__(self) -> str:
        return concat_reprs(
            "To be a string",
            self._length_matcher,
            self._contains_matcher,
            self._starts_with_matcher,
            self._ends_with_matcher,
        )



from typing import Any, Type

from pytest_matchers.matchers import Matcher
from pytest_matchers.utils.matcher_utils import (
    between_matcher,
    is_instance_matcher,
    matches_or_none,
)
from pytest_matchers.utils.repr_utils import concat_reprs


class IsNumber(Matcher):
    def __init__(
        self,
        match_type: Type = None,
        *,
        min_value: float = None,
        max_value: float = None,
        inclusive: bool = None,
        min_inclusive: bool = None,
        max_inclusive: bool = None,
    ):
        self._is_instance_matcher = is_instance_matcher(match_type)
        self._between_matcher = between_matcher(
            min_value,
            max_value,
            inclusive,
            min_inclusive,
            max_inclusive,
        )

    def matches(self, value: Any) -> bool:
        return self._matches_type(value) and self._matches_limit(value)

    def _matches_type(self, value: Any) -> bool:
        if self._is_instance_matcher is None:
            try:
                float(value)
                return True
            except (ValueError, TypeError):
                return False
        return self._is_instance_matcher == value

    def _matches_limit(self, value: Any) -> bool:
        return matches_or_none(self._between_matcher, value)

    def __repr__(self) -> str:
        return concat_reprs(
            "To be a number",
            self._is_instance_matcher,
            self._between_matcher,
        )



from typing import Any, Callable

from pytest_matchers.matchers import (
    Between,
    Contains,
    EndsWith,
    Eq,
    IsInstance,
    Length,
    Matcher,
    StartsWith,
)


def as_matcher(value: Matcher | Any) -> Matcher:
    if isinstance(value, Matcher):
        return value
    return Eq(value)


def as_matcher_or_none(value: Matcher | Any | None) -> Matcher | None:
    if value is None:
        return None
    return as_matcher(value)


def matches_or_none(matcher: Matcher | None, value: Any) -> bool:
    return matcher is None or matcher == value


def partial_matches_or_none(matcher: Matcher | None) -> Callable[[Any], bool]:
    def _partial_matches_or_none(value: Any) -> bool:
        return matches_or_none(matcher, value)

    return _partial_matches_or_none


def is_instance_matcher(match_type: type | None) -> IsInstance | None:
    if match_type is None:
        return None
    return IsInstance(match_type)


def contains_matcher(contains: str | None) -> Contains | None:
    if contains is None:
        return None
    return Contains(contains)


def length_matcher(
    length: int | None,
    min_length: int | None,
    max_length: int | None,
) -> Length | None:
    if length is None and max_length is None and min_length is None:
        return None
    return Length(length, min_length, max_length)


def starts_with_matcher(starts_with: str | None) -> StartsWith | None:
    if starts_with is None:
        return None
    return StartsWith(starts_with)


def ends_with_matcher(ends_with: str | None) -> EndsWith | None:
    if ends_with is None:
        return None
    return EndsWith(ends_with)


def between_matcher(
    min_value: float | None,
    max_value: float | None,
    inclusive: bool | None,
    min_inclusive: bool | None,
    max_inclusive: bool | None,
) -> Between | None:
    if min_value is None and max_value is None:
        return None
    return Between(
        min_value,
        max_value,
        inclusive=inclusive,
        min_inclusive=min_inclusive,
        max_inclusive=max_inclusive,
    )



from typing import Any, Sized

from pytest_matchers.matchers import Matcher


class EndsWith(Matcher):
    def __init__(self, suffix: Sized | str):
        self._suffix = suffix

    def matches(self, value: Any) -> bool:
        if isinstance(value, str):
            try:
                return value.endswith(self._suffix)
            except TypeError:
                return False
        if isinstance(value, list):
            suffix_len = len(self._suffix)
            return value[-suffix_len:] == self._suffix

        return False

    def __repr__(self) -> str:
        return f"To end with {repr(self._suffix)}"

    def concatenated_repr(self) -> str:
        return f"ending with {repr(self._suffix)}"



from typing import Any

from pytest_matchers.matchers import IsInstance, Matcher
from pytest_matchers.utils.matcher_utils import as_matcher
from pytest_matchers.utils.repr_utils import concat_reprs, non_capitalized


class Dict(Matcher):
    def __init__(self, matching: dict = None, *, exclude: list = None):
        self._is_instance_matcher = IsInstance(dict)
        self._matching = matching or {}
        self._exclude = exclude or []

    def matches(self, value: Any) -> bool:
        return (
            self._is_instance_matcher == value
            and not (set(self._exclude) & set(value.keys()))
            and self._matches_values(value)
        )

    def _matches_values(self, value: dict) -> bool:
        for key, matcher in self._matching.items():
            if key not in value or not matcher == value.get(key):
                return False
        return True

    def __repr__(self):
        return concat_reprs("To be a dictionary", self.expectations_repr())

    def expectations_repr(self):
        exclude_repr = f"excluding {', '.join(map(repr, self._exclude))}" if self._exclude else ""
        matching_reprs = [
            f"expecting {repr(key)} {as_matcher(value).concatenated_repr()}"
            for key, value in self._matching.items()
        ]
        return non_capitalized(concat_reprs("", exclude_repr, *matching_reprs))


def dict_matcher(matching: dict | None, exclude: list | None) -> Dict | None:
    if matching is None and exclude is None:
        return None
    return Dict(matching, exclude=exclude)



from typing import Any

from pytest_matchers.matchers import Matcher
from pytest_matchers.utils.matcher_utils import as_matcher, matches_or_none


class HasAttribute(Matcher):
    def __init__(self, attribute_name: str, expected_value: Any = None):
        self._attribute_name = attribute_name
        self._expected_value = expected_value
        self._expected_value_matcher = as_matcher(expected_value) if expected_value else None

    def matches(self, value: Any) -> bool:
        return hasattr(value, self._attribute_name) and matches_or_none(
            self._expected_value_matcher,
            getattr(value, self._attribute_name),
        )

    def __repr__(self) -> str:
        base_repr = f"To have attribute {repr(self._attribute_name)}"
        if self._expected_value_matcher:
            return f"{base_repr} and {self._expected_value_matcher.concatenated_repr()}"
        return base_repr

    def concatenated_repr(self) -> str:
        base_repr = f"with {repr(self._attribute_name)}"
        if self._expected_value:
            if isinstance(self._expected_value, Matcher):
                return f"{base_repr} expecting {self._expected_value.concatenated_repr()}"
            return f"{base_repr} being {repr(self._expected_value)}"
        return base_repr


def has_attribute_matcher(
    attribute: str | None,
    value: Any,
    value_needed: bool = False,
) -> HasAttribute | None:
    if not attribute:
        return None
    if value_needed and value is None:
        return None
    return HasAttribute(attribute, value)



from typing import Any

from pytest_matchers.matchers import Matcher


class IsInstance(Matcher):
    def __init__(self, match_type):
        self._match_type = match_type

    def matches(self, value: Any) -> bool:
        return isinstance(value, self._match_type)

    def __repr__(self) -> str:
        return f"To be instance of '{self._match_type.__name__}'"

    def concatenated_repr(self) -> str:
        return f"of '{self._match_type.__name__}' instance"



from datetime import datetime

from pytest_matchers.matchers import IsString, Matcher
from pytest_matchers.utils.matcher_utils import between_matcher, matches_or_none
from pytest_matchers.utils.repr_utils import concat_reprs


class DatetimeString(Matcher):
    def __init__(
        self,
        expected_format: str,
        *,
        min_value: datetime | str = None,
        max_value: datetime | str = None,
    ):
        self._expected_format = expected_format
        self._between_matcher = between_matcher(
            self._as_datetime(min_value, "min_value"),
            self._as_datetime(max_value, "max_value"),
            None,
            None,
            None,
        )

    def _as_datetime(self, value: datetime | str, value_name: str) -> datetime | None:
        if isinstance(value, str):
            try:
                return datetime.strptime(value, self._expected_format)
            except ValueError as error:
                raise ValueError(
                    f"Invalid datetime string format for {value_name}: {value}"
                ) from error
        if isinstance(value, datetime):
            return datetime.strptime(
                datetime.strftime(value, self._expected_format),
                self._expected_format,
            )
        return None

    def matches(self, value: str) -> bool:
        return IsString() == value and self._matches_format_and_limit(value)

    def _matches_format_and_limit(self, value: str) -> bool:
        try:
            date_time = datetime.strptime(value, self._expected_format)
            return matches_or_none(self._between_matcher, date_time)
        except ValueError:
            return False

    def __repr__(self) -> str:
        return concat_reprs(
            f"To be a datetime string with format {repr(self._expected_format)}",
            self._between_matcher,
        )



from typing import Any

from pytest_matchers.matchers import Matcher


class Anything(Matcher):
    def matches(self, _value: Any) -> bool:
        return True

    def __repr__(self) -> str:
        return "To be anything"



from typing import Any

from pytest_matchers.matchers import Matcher
from pytest_matchers.utils.matcher_utils import as_matcher, as_matcher_or_none


class Case(Matcher):
    def __init__(
        self,
        case_value: Any,
        expectations: dict[Any, Matcher | Any],
        default_expectation: Matcher | Any | None = None,
    ):
        self._case_value = case_value
        self._expectations = expectations
        self._default_expectation = (
            None if default_expectation is None else as_matcher(default_expectation)
        )

    def matches(self, value: Any) -> bool:
        expectation = self._expectation()
        if expectation is None:
            return False
        return value == expectation

    def _expectation(self) -> Matcher | Any | None:
        return as_matcher_or_none(
            self._expectations.get(self._case_value, self._default_expectation)
        )

    def __repr__(self):
        expectation = self._expectation()
        if expectation is None:
            return (
                f"To never match because case value {repr(self._case_value)} is not expected "
                "and the default expectation is not set"
            )
        return f"{repr(expectation)} because case value is {repr(self._case_value)}"



from datetime import datetime

from pytest_matchers.matchers import Matcher
from pytest_matchers.matchers.has_attribute import has_attribute_matcher
from pytest_matchers.utils.matcher_utils import (
    between_matcher,
    is_instance_matcher,
    matches_or_none,
)
from pytest_matchers.utils.repr_utils import concat_reprs


class Datetime(Matcher):  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        min_value: datetime = None,
        max_value: datetime = None,
        year: int = None,
        month: int = None,
        day: int = None,
        hour: int = None,
        minute: int = None,
        second: int = None,
    ):
        self._is_instance_matcher = is_instance_matcher(datetime)
        self._limit_matcher = between_matcher(min_value, max_value, None, None, None)
        self._year_matcher = has_attribute_matcher("year", year, value_needed=True)
        self._month_matcher = has_attribute_matcher("month", month, value_needed=True)
        self._day_matcher = has_attribute_matcher("day", day, value_needed=True)
        self._hour_matcher = has_attribute_matcher("hour", hour, value_needed=True)
        self._minute_matcher = has_attribute_matcher("minute", minute, value_needed=True)
        self._second_matcher = has_attribute_matcher("second", second, value_needed=True)

    def matches(self, value: datetime) -> bool:
        return all(
            [
                matches_or_none(self._is_instance_matcher, value),
                matches_or_none(self._limit_matcher, value),
                matches_or_none(self._year_matcher, value),
                matches_or_none(self._month_matcher, value),
                matches_or_none(self._day_matcher, value),
                matches_or_none(self._hour_matcher, value),
                matches_or_none(self._minute_matcher, value),
                matches_or_none(self._second_matcher, value),
            ]
        )

    def __repr__(self) -> str:
        return concat_reprs(
            "To be a datetime",
            self._limit_matcher,
            self._year_matcher,
            self._month_matcher,
            self._day_matcher,
            self._hour_matcher,
            self._minute_matcher,
            self._second_matcher,
        )



from typing import Any

from pytest_matchers.matchers import Eq, Matcher
from pytest_matchers.utils.repr_utils import concat_reprs, non_capitalized


class Between(Matcher):
    def __init__(
        self,
        min_value: Any,
        max_value: Any,
        inclusive: bool = None,
        min_inclusive: bool = None,
        max_inclusive: bool = None,
    ):
        if min_value is None and max_value is None:
            raise ValueError("At least one of min or max must be specified")
        self._min = min_value
        self._max = max_value
        if inclusive is not None and (min_inclusive is not None or max_inclusive is not None):
            raise ValueError("Cannot specify inclusive and min_inclusive or max_inclusive")
        if inclusive is not None:
            min_inclusive, max_inclusive = inclusive, inclusive
        self._min_inclusive = min_inclusive or min_inclusive is None
        self._max_inclusive = max_inclusive or max_inclusive is None

    def __new__(cls, min_value, max_value, inclusive=None, min_inclusive=None, max_inclusive=None):
        if min_value is not None and min_value == max_value:
            return Eq(min_value)
        return super().__new__(cls)

    def matches(self, value: Any) -> bool:
        try:
            return self._matches_min(value) and self._matches_max(value)
        except TypeError:
            return False

    def _matches_min(self, value: Any) -> bool:
        if self._min is None:
            return True
        if self._min_inclusive:
            return value >= self._min
        return value > self._min

    def _matches_max(self, value: Any) -> bool:
        if self._max is None:
            return True
        if self._max_inclusive:
            return value <= self._max
        return value < self._max

    def _suffix_repr(self) -> str:
        if self._min and self._max:
            if self._min_inclusive and self._max_inclusive:
                return f"between {self._min} and {self._max}"
            if not self._min_inclusive and not self._max_inclusive:
                return f"between {self._min} and {self._max} exclusive"
        min_repr = self._min_repr()
        max_repr = self._max_repr()
        suffix = concat_reprs("", min_repr, max_repr)
        return non_capitalized(suffix)

    def __repr__(self) -> str:
        return f"To be {self._suffix_repr()}"

    def concatenated_repr(self) -> str:
        return self._suffix_repr()

    def _min_repr(self):
        if not self._min:
            return ""
        return (
            f"greater or equal than {self._min}"
            if self._min_inclusive
            else f"greater than {self._min}"
        )

    def _max_repr(self):
        if not self._max:
            return ""
        return (
            f"lower or equal than {self._max}" if self._max_inclusive else f"lower than {self._max}"
        )



from typing import Any

from pytest_matchers.matchers import Matcher


class Eq(Matcher):
    """Why would you want to use this, be serious"""

    def __init__(self, match_value: Any):
        self._match_value = match_value

    def matches(self, value: Any) -> bool:
        return value == self._match_value

    def __repr__(self) -> str:
        return f"To be {repr(self._match_value)}"

    def concatenated_repr(self) -> str:
        return f"equal to {repr(self._match_value)}"



import os
import warnings
from typing import Type


def _warnings_enabled() -> bool:
    return os.environ.get("PYTEST_MATCHERS_WARNINGS", "True") == "True"


def warn(warning: str, category: Type[Warning] = UserWarning) -> None:
    if _warnings_enabled():
        warnings.warn(warning, category)



from typing import Any

from pytest_matchers.matchers import Matcher


class Contains(Matcher):
    def __init__(self, contained_value: Any):
        self._contained_value = contained_value

    def matches(self, value: Any) -> bool:
        try:
            return self._contained_value in value
        except TypeError:
            return False

    def __repr__(self) -> str:
        return f"To contain {repr(self._contained_value)}"

    def concatenated_repr(self) -> str:
        return f"containing {repr(self._contained_value)}"



from typing import Any

from pytest_matchers.matchers import Matcher


class SameValue(Matcher):
    def __init__(self):
        self._matched_value = None

    def matches(self, value: Any) -> bool:
        if self._matched_value is None:
            self._matched_value = value
            return True
        return self._matched_value == value

    def __repr__(self) -> str:
        if self._matched_value is None:
            return "To be the same value"
        return f"To be the same value as {repr(self._matched_value)}"



from typing import Any

from pytest_matchers.matchers import Matcher


class Length(Matcher):
    def __init__(self, length=None, min_length=None, max_length=None):
        if length:
            if min_length or max_length:
                raise ValueError("Cannot specify length with min_length or max_length")
            min_length = length
            max_length = length
        self._min_length = min_length
        self._max_length = max_length

    def matches(self, value: Any) -> bool:
        try:
            value_length = len(value)
        except TypeError:
            return False
        matches_min = self._min_length <= value_length if self._min_length else True
        matches_max = value_length <= self._max_length if self._max_length else True
        return matches_min and matches_max

    def _length_repr(self) -> str:
        if self._min_length is not None and self._max_length is not None:
            if self._min_length == self._max_length:
                return f"of {self._min_length}"
            return f"between {self._min_length} and {self._max_length}"
        if self._min_length is not None:
            return f"greater or equal than {self._min_length}"
        if self._max_length is not None:
            return f"lower or equal than {self._max_length}"
        return ""

    def concatenated_repr(self) -> str:
        length_repr = self._length_repr()
        if length_repr:
            return f"with length {length_repr}"
        return length_repr

    def __repr__(self) -> str:
        length_repr = self._length_repr()
        if length_repr:
            return f"To have length {length_repr}"
        return length_repr



# pylint: disable=import-outside-toplevel
from abc import ABC, abstractmethod
from typing import Any


class Matcher(ABC):
    def __eq__(self, other: Any) -> bool:
        return self.matches(other)

    def __and__(self, other: "Matcher") -> "Matcher":
        from pytest_matchers.matchers import And

        return And(self, other)

    def __or__(self, other: "Matcher") -> "Matcher":
        from pytest_matchers.matchers import Or

        return Or(self, other)

    def __invert__(self) -> "Matcher":
        from pytest_matchers.matchers import Not

        return Not(self)

    @abstractmethod
    def matches(self, value: Any) -> bool:
        pass

    def concatenated_repr(self) -> str:
        from pytest_matchers.utils.repr_utils import non_capitalized

        return non_capitalized(repr(self))



from pytest_matchers.matchers import Matcher


def _repr(extra_repr: str | Matcher | None) -> str:
    if isinstance(extra_repr, Matcher):
        return extra_repr.concatenated_repr()
    return extra_repr


def concat_matcher_repr(matcher: Matcher | None) -> str:
    return _repr(matcher)


def concat_reprs(
    base_repr: str | Matcher | None,
    *extra_reprs: str | Matcher | None,
    separator: str = "and",
) -> str:
    base_repr = str(base_repr) if base_repr else ""
    extra_repr = f" {separator} ".join(
        [_repr(extra_repr) for extra_repr in extra_reprs if extra_repr]
    )
    if not base_repr and extra_repr:
        return capitalized(extra_repr)
    if extra_repr:
        return f"{base_repr} {extra_repr}"
    return base_repr


def capitalized(string: str):
    if not string:
        return string
    return string[0].upper() + string[1:]


def non_capitalized(string: str):
    if not string:
        return string
    return string[0].lower() + string[1:]



from typing import Any, Callable, Type

from pytest_matchers.matchers import (
    Anything,
    Between,
    Case,
    Datetime,
    DatetimeString,
    Dict,
    DifferentValue,
    HasAttribute,
    If,
    IsInstance,
    IsList,
    IsNumber,
    IsString,
    JSON,
    Matcher,
    Or,
    SameValue,
)


def anything() -> Anything:
    return Anything()


def is_instance(match_type: Type) -> IsInstance:
    return IsInstance(match_type)


def is_list(match_type: Type = None, **kwargs) -> IsList | IsInstance:
    if match_type is None and not kwargs:
        return is_instance(list)
    return IsList(match_type, **kwargs)


def is_string(**kwargs) -> IsString:
    return IsString(**kwargs)


def is_number(match_type: Type = None, **kwargs) -> IsNumber:
    return IsNumber(match_type, **kwargs)


def one_of(*values: Matcher | Any) -> Or:
    return Or(*values)


def has_attribute(attribute_name: str, expected_value: Any = None) -> HasAttribute:
    return HasAttribute(attribute_name, expected_value)


def between(
    min_value: Any,
    max_value: Any,
    *,
    inclusive: bool = None,
    min_inclusive: bool = None,
    max_inclusive: bool = None,
) -> Between:
    return Between(min_value, max_value, inclusive, min_inclusive, max_inclusive)


def is_datetime(
    min_value: Any = None,
    max_value: Any = None,
    year: int = None,
    month: int = None,
    day: int = None,
    hour: int = None,
    minute: int = None,
    second: int = None,
) -> Datetime:
    return Datetime(min_value, max_value, year, month, day, hour, minute, second)


def is_datetime_string(
    expected_format: str,
    *,
    min_value: Any = None,
    max_value: Any = None,
) -> DatetimeString:
    return DatetimeString(expected_format, min_value=min_value, max_value=max_value)


def same_value() -> SameValue:
    return SameValue()


def different_value() -> DifferentValue:
    return DifferentValue()


def if_true(
    condition: Callable | bool | Any,
    then: Matcher | Any = None,
    or_else: Matcher | Any = None,
) -> If:
    return If(condition, then, or_else)


def if_false(
    condition: Callable | bool | Any,
    then: Matcher | Any = None,
    or_else: Matcher | Any = None,
) -> If:
    return if_true(condition, or_else, then)  # pylint: disable=arguments-out-of-order


def case(
    case_value: Any,
    expectations: dict[Any, Matcher | Any],
    default_expectation: Matcher | Any | None = None,
) -> Case:
    return Case(case_value, expectations, default_expectation)


def is_dict(matching: dict = None, *, exclude: list[Any] = None) -> Dict:
    return Dict(matching, exclude=exclude)


def is_json(matching: dict = None, *, exclude: list[Any] = None) -> JSON:
    return JSON(matching, exclude=exclude)



from typing import Any, Callable, Type

from pytest_matchers.matchers import IsInstance, Matcher
from pytest_matchers.utils.matcher_utils import (
    is_instance_matcher,
    length_matcher,
    matches_or_none,
    partial_matches_or_none,
)
from pytest_matchers.utils.repr_utils import concat_reprs


class IsList(Matcher):
    def __init__(
        self,
        match_type: Type = None,
        *,
        length: int = None,
        min_length: int = None,
        max_length: int = None,
    ):
        self._is_instance_matcher = is_instance_matcher(match_type)
        self._length_matcher = length_matcher(length, min_length, max_length)

    def matches(self, value: Any) -> bool:
        return self._matches_type(value) and matches_or_none(self._length_matcher, value)

    def _matches_type(self, value: Any) -> bool:
        return IsInstance(list) == value and all(map(self._type_matcher(), value))

    def _type_matcher(self) -> Callable[[Any], bool]:
        return partial_matches_or_none(self._is_instance_matcher)

    def __repr__(self) -> str:
        return concat_reprs(
            "To be a list",
            self._is_instance_matcher,
            self._length_matcher,
        )



from typing import Any, Sized

from pytest_matchers.matchers import Matcher


class StartsWith(Matcher):
    def __init__(self, prefix: Sized | str):
        self._prefix = prefix

    def matches(self, value: Any) -> bool:
        if isinstance(value, str):
            try:
                return value.startswith(self._prefix)
            except TypeError:
                return False
        if isinstance(value, list):
            return value[: len(self._prefix)] == self._prefix

        return False

    def __repr__(self) -> str:
        return f"To start with {repr(self._prefix)}"

    def concatenated_repr(self) -> str:
        return f"starting with {repr(self._prefix)}"
